import numpy as np
from .kernels import GWRKernel, GTWRKernel
from scipy import linalg

def _compute_betas_gwr(y, x, wi):
    """
    compute MLE coefficients using iwls routine

    Methods: p189, Iteratively (Re)weighted Least Squares (IWLS),
    Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.
    """
    xT = (x * wi).T
    xtx = np.dot(xT, x)
    xtx_inv_xt = linalg.solve(xtx, xT)
    betas = np.dot(xtx_inv_xt, y)
    return betas, xtx_inv_xt


class GWR(object):
    
    def __init__(self, coords, y, X, bw, kernel = 'gaussian', 
                 fixed = False, constant = True):
        
        self.coords = coords
        self.y = y
        self.n = X.shape[0]
        self.bw = bw
        self.kernel = kernel
        self.fixed = fixed
        self.constant = constant
        if self.constant:
            self.X = np.hstack([np.ones((self.n, 1)), X])
        else:
            self.X = X
        self.k = self.X.shape[1]
            
    def _build_wi(self, i, bw):

        try:
            wi = GWRKernel(i, self.coords, bw, fixed=self.fixed,
                        function=self.kernel).kernel
        except BaseException:
            raise  # TypeError('Unsupported kernel function  ', kernel)
            
        return wi
    
    def _local_fit(self, i, final = True):
        
        wi = self._build_wi(i, self.bw).reshape(-1, 1)
        betas, xtx_inv_xt = _compute_betas_gwr(self.y, self.X, wi)
        predy = np.dot(self.X[i], betas)[0]
        resid = self.y[i] - predy
        influ = np.dot(self.X[i], xtx_inv_xt[:, i])
        if not final:
            return resid * resid, influ
        else:
            Si = np.dot(self.X[i], xtx_inv_xt).reshape(-1)
            CCT = np.diag(np.dot(xtx_inv_xt, xtx_inv_xt.T)).reshape(-1)
            Si2 = np.sum(Si**2)
            return influ, resid, predy, betas.reshape(-1), CCT, Si2
    
    def fit(self, final = True):
        
        if not final:
            RSS = 0
            tr_S = 0
            aa = 0
            for i in range(self.n):
                err2, hat = self._local_fit(i, final = False)
                aa += err2 / ((1 - hat) ** 2)
                RSS += err2
                tr_S += hat
            llf = -np.log(RSS) * self.n / \
                2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
            return Notfinal(float(RSS), tr_S, float(llf), float(aa), self.n)
        else:   
            influ, resid, predy = np.empty((self.n, 1)), \
                np.empty((self.n, 1)), np.empty((self.n, 1))
            betas, CCT = np.empty((self.n, self.k)), np.empty((self.n, self.k))
            tr_STS = 0
            for i in range(self.n):
                influi, residi, predyi, betasi, CCTi, tr_STSi = self._local_fit(i)
                influ[i] = influi
                resid[i] = residi
                predy[i] = predyi
                betas[i] = betasi
                CCT[i] = CCTi                 
                tr_STS += tr_STSi        
            return Final(self.coords, None, self.y, self.X, self.bw, None, 
                         self.kernel, self.fixed, self.constant, influ, resid, predy, 
                        betas, CCT, tr_STS, model = 'GWR')
        
        

class TWR(object):
    
    def __init__(self, t, y, X, h, kernel = 'gaussian', 
                 fixed = False, constant = True):
        
        self.t = t
        self.y = y
        self.n = X.shape[0]
        self.h = h
        self.kernel = kernel
        self.fixed = fixed
        self.constant = constant
        if self.constant:
            self.X = np.hstack([np.ones((self.n, 1)), X])
        else:
            self.X = X
        self.k = self.X.shape[1]
            
    def _build_wi(self, i, h):

        try:
            wi = GWRKernel(i, self.t, h, fixed=self.fixed,
                        function=self.kernel).kernel
        except BaseException:
            raise  # TypeError('Unsupported kernel function  ', kernel)
            
        return wi
    
    def _local_fit(self, i, final = True):
        
        wi = self._build_wi(i, self.h).reshape(-1, 1)
        X_derivative = self.X * (self.t-self.t[i])
        X_new = np.hstack([self.X, X_derivative])
        xT = (X_new * wi).T
        xtx_inv_xt = np.dot(np.linalg.inv(np.dot(xT, X_new)), xT)
        xstack = np.hstack([self.X[i].reshape(1,self.k),np.zeros((1,self.k))])
        predy = (np.dot(np.dot(xstack, xtx_inv_xt), self.y))[0] 
        resid = self.y[i] - predy
        influ = np.dot(xstack, xtx_inv_xt[:,i])[0]  
        if not final:
            return resid * resid, influ
        else:
            betas = np.dot(xtx_inv_xt, self.y)[:self.k]
            zeros = np.zeros((1,self.k))
            Si = np.dot(np.hstack([self.X[i].reshape(1,self.k),zeros]), xtx_inv_xt).reshape(-1)
            Si2 = np.sum(Si**2)
            return influ, resid, predy, betas.reshape(-1), Si2
    
    def fit(self, final = True):
        
        if not final:
            RSS = 0
            tr_S = 0
            aa = 0
            for i in range(self.n):
                err2, hat = self._local_fit(i, final = False)
                aa += err2 / ((1 - hat) ** 2)
                RSS += err2
                tr_S += hat
            llf = -np.log(RSS) * self.n / \
                2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
            return Notfinal(float(RSS), tr_S, float(llf), float(aa), self.n)
        else:   
            influ, resid, predy = np.empty((self.n, 1)), \
                np.empty((self.n, 1)), np.empty((self.n, 1))
            betas = np.empty((self.n, self.k))
            tr_STS = 0
            for i in range(self.n):
                influi, residi, predyi, betasi, tr_STSi = self._local_fit(i)
                influ[i] = influi
                resid[i] = residi
                predy[i] = predyi
                betas[i] = betasi                
                tr_STS += tr_STSi        
            return Final(self.t, None, self.y, self.X, self.h, None, 
                         self.kernel, self.fixed, self.constant, influ, resid, predy, 
                        betas, None, tr_STS, model = 'TWR')
       

class GTWR(object):

    def __init__(self, coords, t, y, X, bw, tau, kernel = 'gaussian', 
                 fixed = False, constant = True):
        self.coords = coords
        self.t = t
        self.y = y
        self.n = X.shape[0]
        self.bw = bw
        self.tau = tau
        self.kernel = kernel
        self.fixed = fixed
        self.constant = constant
        if self.constant:
            self.X = np.hstack([np.ones((self.n, 1)), X])
        else:
            self.X = X
        self.k = self.X.shape[1]
        self.bw_s = self.bw
        self.bw_t = np.sqrt(self.bw**2 / self.tau)
        
    def _build_wi(self, i, bw, tau):

        try:
            wi = GTWRKernel(i, self.coords, self.t, bw, tau, fixed=self.fixed,
                        function=self.kernel).kernel
        except BaseException:
            raise  # TypeError('Unsupported kernel function  ', kernel)
            
        return wi

    def _local_fit(self, i, final = True, multi = False):
        
        wi = self._build_wi(i, self.bw, self.tau).reshape(-1, 1)
        betas, xtx_inv_xt = _compute_betas_gwr(self.y, self.X, wi)
        predy = np.dot(self.X[i], betas)[0]
        resid = self.y[i] - predy
        influ = np.dot(self.X[i], xtx_inv_xt[:, i])
        if not final:
            return resid * resid, influ
        else:
            Si = np.dot(self.X[i], xtx_inv_xt).reshape(-1)
            CCT = np.diag(np.dot(xtx_inv_xt, xtx_inv_xt.T)).reshape(-1)
            Si2 = np.sum(Si**2)
            return influ, resid, predy, betas.reshape(-1), CCT, Si2 
        
    def fit(self, final = True, mpi = False):
        
        if mpi:
            from mpi4py import MPI
            import math
            comm = MPI.COMM_WORLD
            size = comm.Get_size()
            rank = comm.Get_rank()
            iter = np.arange(self.n)
            m = int(math.ceil(float(len(iter)) / size))
            x_chunk = iter[rank * m:(rank + 1) * m]
        if not final: 
            if mpi:
                RSS = 0
                tr_S = 0
                aa = 0
                for i in x_chunk:       
                    err2, hat = self._local_fit(i, final = False)
                    aa += err2 / ((1 - hat) ** 2)
                    RSS += err2
                    tr_S += hat
                aa_list = comm.gather(aa, root=0)
                RSS_list = comm.gather(RSS, root=0)
                trS_list = comm.gather(tr_S, root=0)
                if rank == 0:
                    RSS = sum(RSS_list)
                    tr_S = sum(trS_list)
                    aa = sum(aa_list)
                    llf = -np.log(RSS) * self.n / \
                    2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
                    return Notfinal(float(RSS), tr_S, float(llf), float(aa), self.n)
            else:
                RSS = 0
                tr_S = 0
                aa = 0
                for i in range(self.n):
                    err2, hat = self._local_fit(i, final = False)
                    aa += err2 / ((1 - hat) ** 2)
                    RSS += err2
                    tr_S += hat
                llf = -np.log(RSS) * self.n / \
                    2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
                return Notfinal(float(RSS), tr_S, float(llf), float(aa), self.n)
        else:
            if mpi:
                n_chunk = x_chunk.shape[0]
                influ, resid, predy = np.empty((n_chunk, 1)), \
                np.empty((n_chunk, 1)), np.empty((n_chunk, 1))
                betas, CCT = np.empty((n_chunk, self.k)), np.empty((n_chunk, self.k))
                tr_STS = 0
                pos = 0
                for i in x_chunk:
                    influi, residi, predyi, betasi, CCTi, tr_STSi = self._local_fit(i)
                    influ[pos] = influi
                    resid[pos] = residi
                    predy[pos] = predyi
                    betas[pos] = betasi
                    CCT[pos] = CCTi                   
                    tr_STS += tr_STSi
                    pos += 1
                influ_list = comm.gather(influ, root=0)
                resid_list = comm.gather(resid, root=0)
                predy_list = comm.gather(predy, root=0)
                betas_list = comm.gather(betas, root=0)
                CCT_list = comm.gather(CCT, root=0)
                tr_STS_list = comm.gather(tr_STS, root=0)   
                if rank == 0:
                    influ = np.vstack(influ_list)
                    resid = np.vstack(resid_list)
                    predy = np.vstack(predy_list)
                    betas = np.vstack(betas_list)
                    CCT = np.vstack(CCT_list)
                    tr_STS = sum(tr_STS_list)                        
                    return Final(self.coords, self.t, self.y, self.X, self.bw, 
                                 self.tau, self.kernel, self.fixed, self.constant, 
                                 influ, resid, predy, betas, CCT, tr_STS, model = 'GTWR')
                            
            else:   
                influ, resid, predy = np.empty((self.n, 1)), \
                    np.empty((self.n, 1)), np.empty((self.n, 1))
                betas, CCT = np.empty((self.n, self.k)), np.empty((self.n, self.k))
                tr_STS = 0
                for i in range(self.n):
                    influi, residi, predyi, betasi, CCTi, tr_STSi = self._local_fit(i)
                    influ[i] = influi
                    resid[i] = residi
                    predy[i] = predyi
                    betas[i] = betasi
                    CCT[i] = CCTi                   
                    tr_STS += tr_STSi           
                return Final(self.coords, self.t, self.y, self.X, self.bw, 
                             self.tau, self.kernel, self.fixed, self.constant, 
                             influ, resid, predy, betas, CCT, tr_STS, model = 'GTWR')
        
        
        
        
        
        
    
class Notfinal(object):
    
     def __init__(self, RSS, tr_S, llf, aa, n):
         self.RSS = RSS
         self.tr_S = tr_S
         self.llf = llf
         self.aa = aa
         self.n = n      
    
class Final(object):
   
    def __init__(self, coords, t, y, X, bw, tau, kernel, fixed, constant,
                 influ, resid, predy, betas, CCT, tr_STS, model):
        if model == 'GWR':
            GWR.__init__(self, coords, y, X, bw,  kernel, fixed, constant=False)
        if model == 'GTWR':
            GTWR.__init__(self, coords, t, y, X, bw, tau, kernel, fixed, constant=False)
        if model == 'TWR':
            TWR.__init__(self, t, y, X, bw,  kernel, fixed, constant=False)
        self.influ = influ
        self.resid = resid
        self.predy = predy
        self.betas = betas
        self.tr_S = np.sum(influ)
        self.ENP = self.tr_S
        self.tr_STS = tr_STS
        self.TSS = np.sum((y - np.mean(y))**2)
        self.RSS = np.sum(resid**2)
        self.sigma2 = self.RSS / (self.n - self.tr_S)
        if CCT is not None:
            self.CCT = CCT * self.sigma2
            self.bse = np.sqrt(self.CCT)
            self.tvalues = self.betas / self.bse
        else:
            self.CCT = None
            self.bse = None
            self.tvalues = None
        self.std_res = self.resid / (np.sqrt(self.sigma2 * (1.0 - self.influ)))      
        self.cooksD = self.std_res**2 * self.influ / \
            (self.tr_S * (1.0 - self.influ))    
        self.df_model = self.n - self.tr_S
        self.df_resid = self.n - 2.0 * self.tr_S + self.tr_STS
        self.R2 = 1 - self.RSS / self.TSS
        self.adj_R2 = 1 - (1 - self.R2) * (self.n - 1) / (self.n - self.ENP - 1)
        self.llf = -np.log(self.RSS) * self.n / \
                2 - (1 + np.log(np.pi / self.n * 2)) * self.n / 2
        self.aic = -2.0 * self.llf + 2.0 * (self.tr_S + 1)
        self.aicc = self.aic + 2.0 * self.tr_S * (self.tr_S + 1.0) / \
                  (self.n - self.tr_S - 1.0)
        self.bic = -2.0 * self.llf + (self.k + 1) * \
            np.log(self.n)    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    


import numpy as np
from .diagnosis import get_AICc, get_AIC, get_BIC, get_CV
from scipy.spatial.distance import pdist
from .model import GWR, TWR, GTWR
from .search import golden_section, twostep_golden_section

getDiag = {'AICc': get_AICc, 'AIC': get_AIC, 'BIC': get_BIC, 'CV': get_CV}

delta = 0.38197
class search_GWRparameter(object):
    
    def __init__(self, coords, y, X, kernel = 'exponential', fixed = False, constant=True):
        self.coords = coords
        self.y = y
        self.X = X
        self.n = X.shape[0]
        self.constant = constant
        if constant:
            self.k = X.shape[1] + 1
        else:
            self.k = X.shape[1]
        self.kernel = kernel
        self.fixed = fixed
        self.int_score = not self.fixed
        
    def search(self, criterion='AICc', bw_min = None, bw_max = None, tol=1.0e-6, bwdecimal = 0, 
               max_iter = 200, verbose = False):
        
        self.criterion = criterion
        self.bw_min = bw_min
        self.bw_max = bw_max
        self.tol = tol
        self.bwdecimal = bwdecimal
        self.max_iter = max_iter
        self.verbose = verbose
        
        self._bw()
        return self.bw
    
    def _bw(self):
        
        gwr_func = lambda bw : getDiag[self.criterion](GWR(
                self.coords, self.y, self.X, bw, kernel=self.kernel, 
                fixed=self.fixed, constant=self.constant).fit(final = False))
        
        self.bw_min, self.bw_max = self._init_section(self.X, self.coords, self.constant)
        self.bw = golden_section(self.bw_min, self.bw_max, delta, self.bwdecimal, 
                                 gwr_func, self.tol,self.max_iter, self.verbose)
    
    def _init_section(self, X, coords, constant):

        if len(X) > 0:
            n_glob = X.shape[1]
        else:
            n_glob = 0
        if constant:
            n_vars = n_glob + 1
        else:
            n_vars = n_glob
        n = np.array(coords).shape[0]

        if self.int_score:
            a = 40 + 2 * n_vars
            c = n
        else:
            sq_dists = pdist(coords)
            a = np.min(sq_dists) / 2.0
            c = np.max(sq_dists) 
        if self.bw_min is not None:
            a = self.bw_min
        if self.bw_max is not None:
            c = self.bw_max
            
        return a, c
    
class search_TWRparameter(object):
    
    def __init__(self, t, y, X, kernel = 'exponential', fixed = False, constant=True):
        self.t = t
        self.y = y
        self.X = X
        self.n = X.shape[0]
        self.constant = constant
        if constant:
            self.k = X.shape[1] + 1
        else:
            self.k = X.shape[1]
        self.kernel = kernel
        self.fixed = fixed
        
    def search(self, criterion='AICc', h_min = None, h_max = None, tol=1.0e-6, hdecimal = 0, 
               max_iter = 200, verbose = False):
        
        self.criterion = criterion
        self.h_min = h_min
        self.h_max = h_max
        self.tol = tol
        self.hdecimal = hdecimal
        self.max_iter = max_iter
        self.verbose = verbose
        
        self._h()
        return self.h
    
    def _h(self):
        
        twr_func = lambda h : getDiag[self.criterion](TWR(
                self.t, self.y, self.X, h, kernel=self.kernel, 
                fixed=self.fixed, constant=self.constant).fit(final = False))
        
        self.h_min, self.h_max = self._init_section(self.X, self.t, self.constant)
        self.h = golden_section(self.h_min, self.h_max, delta, self.hdecimal, 
                                 twr_func, self.tol,self.max_iter, self.verbose)
    
    def _init_section(self, X, t, constant):

        sq_dists = pdist(t)
        a = np.min(sq_dists) / 2.0
        c = np.max(sq_dists) 
        if self.h_min is not None:
            a = self.h_min
        if self.h_max is not None:
            c = self.h_max
        return a, c
        
class search_GTWRparameter(object):
    
    def __init__(self, coords, t, y, X, kernel = 'exponential', fixed = False, constant=True):
        self.coords = coords
        self.t = t
        self.y = y
        self.X = X
        self.n = X.shape[0]
        self.constant = constant
        if constant:
            self.k = X.shape[1] + 1
        else:
            self.k = X.shape[1]
        self.kernel = kernel
        self.fixed = fixed
        self.int_score = not self.fixed
        
    def search(self, criterion='AICc', bw_min = None, bw_max = None, tau_min = None, 
               tau_max = None, tol=1.0e-6, bwdecimal = 0, taudecimal = 1, max_iter = 200,
               verbose = False, mpi = False):
        
        self.criterion = criterion
        self.bw_min = bw_min
        self.bw_max = bw_max
        self.tau_min = tau_min
        self.tau_max = tau_max
        self.tol = tol
        self.bwdecimal = bwdecimal
        self.taudecimal = taudecimal
        self.max_iter = max_iter
        self.verbose = verbose
        self.mpi = mpi
        
        self._bwtau()
        return self.bw, self.tau
    
    def _bwtau(self):
        if self.mpi:
            gtwr_func = lambda bw, tau: getDiag[self.criterion](GTWR(
                self.coords, self.t, self.y, self.X, bw, tau, kernel=self.kernel, 
                fixed=self.fixed, constant=self.constant).fit(final = False,mpi=True),mpi=True)
            self.bw_min, self.bw_max, self.tau_min, self.tau_max = \
                self._init_section(self.X, self.coords, self.constant)
            self.bw, self.tau = twostep_golden_section(self.bw_min, self.bw_max, 
                self.tau_min, self.tau_max, delta, gtwr_func, self.tol, self.max_iter, 
                self.bwdecimal, self.taudecimal, self.verbose, self.mpi)
        else:    
            gtwr_func = lambda bw, tau: getDiag[self.criterion](GTWR(
                self.coords, self.t, self.y, self.X, bw, tau, kernel=self.kernel, 
                fixed=self.fixed, constant=self.constant).fit(final = False))
            self.bw_min, self.bw_max, self.tau_min, self.tau_max = \
                self._init_section(self.X, self.coords, self.constant)
            self.bw, self.tau = twostep_golden_section(self.bw_min, self.bw_max, 
                self.tau_min, self.tau_max, delta, gtwr_func, self.tol,
                self.max_iter, self.bwdecimal, self.taudecimal, self.verbose)
            
    def _init_section(self, X, coords, constant):
        if len(X) > 0:
            n_glob = X.shape[1]
        else:
            n_glob = 0
        if constant:
            n_vars = n_glob + 1
        else:
            n_vars = n_glob
        n = np.array(coords).shape[0]

        if self.int_score:
            a = 40 + 2 * n_vars
            c = n
        else:
            sq_dists = pdist(coords)
            a = np.min(sq_dists) / 2.0
            c = np.max(sq_dists) 

        if self.bw_min is not None:
            a = self.bw_min
        if self.bw_max is not None:
            c = self.bw_max
        
        if self.tau_min is not None:
            A = self.tau_min
        else:
            A = 0
        if self.tau_max is not None:
            C = self.tau_max
        else:
            C = 4       
        return a, c, A, C

        

        
        
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    


import numpy as np
from .kernels import GTWRKernel, GWRKernel
from scipy import linalg
from .model import _compute_betas_gwr
from scipy.stats import f

class test(object):
    
    def __init__(self, coords, t, y, X, bw_GTWR, tau_GTWR, kernel_GTWR = 'gaussian', 
                 fixed_GTWR = False, constant = True):
        
        self.coords = coords
        self.t = t
        self.y = y
        self.n = X.shape[0]
        self.bw_GTWR = bw_GTWR
        self.tau_GTWR = tau_GTWR
        self.kernel_GTWR = kernel_GTWR
        self.fixed_GTWR = fixed_GTWR
        self.constant = constant
        if self.constant:
            self.X = np.hstack([np.ones((self.n, 1)), X])
        else:
            self.X = X
        self.k = self.X.shape[1]
        self.RSS_GTWR = 0
        S = np.empty((self.n,self.n))
        for i in range(self.n):
            wi = GTWRKernel(i, self.coords, self.t, self.bw_GTWR, self.tau_GTWR,
                           self.fixed_GTWR, function = self.kernel_GTWR).kernel.reshape(-1,1)
            wi[wi <= 1e-6] = 1e-6
            betas, xtx_inv_xt = _compute_betas_gwr(self.y, self.X, wi)
            predy = np.dot(self.X[i], betas)[0]
            resid = self.y[i] - predy
            self.RSS_GTWR += float(resid ** 2)
            S[i] = np.dot(self.X[i], xtx_inv_xt)
        self.Rs = np.dot((np.eye(self.n)-S).T, (np.eye(self.n)-S))
        del S
        self.trRs = 0
        trRsRs = 0
        for i in range(self.n):
            self.trRs += self.Rs[i,i]
            trRsRs += float(self.Rs[i,:] @ self.Rs[:,i])
        self.r_s = self.trRs ** 2 / trRsRs
       
    def spacialtimetest(self,):
        
        H = self.X @ np.linalg.inv(self.X.T @ self.X) @ self.X.T
        Rh = (np.eye(self.n)-H).T @ (np.eye(self.n)-H)
        del H
        RSS_OLS = (self.y.T @ Rh @ self.y)[0][0]
        Rhs = Rh - self.Rs
        del Rh
        trRhs = 0
        trRhsRhs = 0
        for i in range(self.n):
            trRhs += Rhs[i,i]
            trRhsRhs += float(Rhs[i,:] @ Rhs[:,i])
        r_h_s = trRhs**2 / trRhsRhs
        gttest = (RSS_OLS - self.RSS_GTWR) / self.RSS_GTWR * self.trRs/trRhs        
        gtF = f.sf(gttest,r_h_s,self.r_s)
        return gttest, gtF, r_h_s, self.r_s
    
    def spacialtest(self, h, kernel_TWR = 'gaussian', fixed_TWR = False):
        
        RSS_TWR = 0
        M = np.empty((self.n,self.n))
        for i in range(self.n):
            wi = GWRKernel(i, self.t, h, fixed=fixed_TWR,
                        function=kernel_TWR).kernel.reshape(-1,1)
            X_derivative = self.X * (self.t-self.t[i])
            X_new = np.hstack([self.X, X_derivative])
            xT = (X_new * wi).T
            xtx_inv_xt = np.dot(np.linalg.inv(np.dot(xT, X_new)), xT)
            xstack = np.hstack([self.X[i].reshape(1,self.k),np.zeros((1,self.k))])
            predy = (np.dot(np.dot(xstack, xtx_inv_xt), self.y))[0] 
            resid = self.y[i] - predy
            RSS_TWR += float(resid ** 2)
            M[i] = np.dot(xstack, xtx_inv_xt)
        Rm = (np.eye(self.n)-M).T @ (np.eye(self.n)-M)
        del M
        Rms = Rm - self.Rs
        del Rm
        trRms = 0
        trRmsRms = 0
        for i in range(self.n):
            trRms += Rms[i,i]
            trRmsRms += float(Rms[i,:] @ Rms[:,i])
        r_m_s = trRms**2 / trRmsRms
        gtest = (RSS_TWR - self.RSS_GTWR) / self.RSS_GTWR * self.trRs/trRms        
        gF = f.sf(gtest,r_m_s,self.r_s)
        return gtest, gF, r_m_s, self.r_s
    
    def timetest(self, bw_GWR, kernel_GWR = 'gaussian', fixed_GWR = False):
        
        RSS_GWR = 0
        L = np.empty((self.n,self.n))
        for i in range(self.n):
            wi = GWRKernel(i, self.coords, bw_GWR, fixed=fixed_GWR,
                        function=kernel_GWR).kernel.reshape(-1,1)
            betas, xtx_inv_xt = _compute_betas_gwr(self.y, self.X, wi)
            predy = np.dot(self.X[i], betas)[0]
            resid = self.y[i] - predy
            RSS_GWR += float(resid ** 2)
            L[i] = np.dot(self.X[i], xtx_inv_xt)
        Rl = (np.eye(self.n)-L).T @ (np.eye(self.n)-L)
        del L
        Rls = Rl - self.Rs
        del Rl
        trRls = 0
        trRlsRls = 0
        for i in range(self.n):
            trRls += Rls[i,i]
            trRlsRls += float(Rls[i,:] @ Rls[:,i])
        r_l_s = trRls**2 / trRlsRls
        ttest = (RSS_GWR - self.RSS_GTWR) / self.RSS_GTWR * self.trRs/trRls                
        tF = f.sf(ttest,r_l_s,self.r_s)
        return ttest, tF, r_l_s, self.r_s
        
        
        

        
        
            
            
        
        


import numpy as np

def golden_section(a, c, delta, decimal, function, tol, max_iter, verbose=False):
    b = a + delta * np.abs(c - a)
    d = c - delta * np.abs(c - a)
    diff = 1.0e9
    iters = 0
    dict = {}
    while np.abs(diff) > tol and iters < max_iter:
        iters += 1
        b = np.round(b, decimal)
        d = np.round(d, decimal)

        if b in dict:
            score_b = dict[b]
        else:
            score_b = function(b)
            dict[b] = score_b

        if d in dict:
            score_d = dict[d]
        else:
            score_d = function(d)
            dict[d] = score_d

        if score_b <= score_d:
            opt_val = b
            opt_score = score_b
            c = d
            d = b
            b = a + delta * np.abs(c - a)

        else:
            opt_val = d
            opt_val = score_d
            a = b
            b = d
            d = c - delta * np.abs(c - a)

        
        opt_val = np.round(opt_val, decimal)        
        diff = score_b - score_d
        if verbose:
            print('bw:', opt_val, ', score:', np.round(opt_score,2))

    return opt_val



def onestep_golden_section(A, C, x, delta, taudecimal, function, tol, mpi=False):
    iters = 0
    dict = {}
    diff = 1e9
    opt_score = None
    opt_tau = None
    B = A + delta * np.abs(C - A)
    D = C - delta * np.abs(C - A)
    while np.abs(diff) > tol and iters < 200:
        iters += 1
        B = np.round(B, taudecimal)
        D = np.round(D, taudecimal)
        if B in dict:
            score_B = dict[B]
        else:
            score_B = function(x, B)
            dict[B] = score_B

        if D in dict:
            score_D = dict[D]
        else:
            score_D = function(x, D)
            dict[D] = score_D
        if mpi:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
            if rank == 0:
                if score_B <= score_D:
                    opt_score = score_B
                    opt_tau = B
                    C = D
                    D = B
                    B = A + delta * np.abs(C - A)
                else:
                    opt_score = score_D
                    opt_tau = D
                    A = B
                    B = D
                    D = C - delta * np.abs(C - A)
                diff = score_B - score_D
                opt_tau = np.round(opt_tau, taudecimal)
            B = comm.bcast(B, root=0)
            D = comm.bcast(D, root=0)
            diff = comm.bcast(diff, root=0)
            opt_score = comm.bcast(opt_score, root=0)
            opt_tau = comm.bcast(opt_tau, root=0)
        else:
            if score_B <= score_D:
                opt_score = score_B
                opt_tau = B
                C = D
                D = B
                B = A + delta * np.abs(C - A)
            else:
                opt_score = score_D
                opt_tau = D
                A = B
                B = D
                D = C - delta * np.abs(C - A)
            diff = score_B - score_D     
    return opt_tau, opt_score

def twostep_golden_section(a, c, A, C, delta, function, 
                        tol, max_iter, bwdecimal, taudecimal, verbose = False, mpi = False):
    b = a + delta * np.abs(c - a)
    d = c - delta * np.abs(c - a)
    opt_score = None
    opt_bw = None
    opt_tau = None
    diff = 1e9
    dict = {}
    iters = 0
    while np.abs(diff) > tol and iters < 200:       
        iters += 1
        b = np.round(b, bwdecimal)
        d = np.round(d, bwdecimal)
        if b in dict:
            score_b = dict[b]
        else:
            if mpi:
                tau_b, score_b = onestep_golden_section(A, C, b, delta, taudecimal, function, 
                                            tol, mpi=True)
            else:
                tau_b, score_b = onestep_golden_section(A, C, b, delta, taudecimal, function, 
                                            tol)
            dict[b] = score_b
        if d in dict:
            score_d = dict[d]
        else:
            if mpi:
                tau_d, score_d = onestep_golden_section(A, C, d, delta, taudecimal, function, 
                                            tol, mpi=True)
            else:
                tau_d, score_d = onestep_golden_section(A, C, d, delta, taudecimal, function, 
                                            tol)
            dict[d] = score_d
        if mpi:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
            if rank == 0:
                if score_b <= score_d:
                    opt_score = score_b
                    opt_bw = b
                    opt_tau = tau_b
                    c = d
                    d = b
                    b = a + delta * np.abs(c - a)
                else:
                    opt_score = score_d
                    opt_bw = d
                    opt_tau = tau_d
                    a = b
                    b = d
                    d = c - delta * np.abs(c - a) 
                diff = score_b - score_d
                opt_tau = np.round(opt_tau, taudecimal)
                opt_bw = np.round(opt_bw, bwdecimal)
                if verbose:            
                    print('bw: ', opt_bw, ', tau: ', opt_tau, ', score: ', opt_score)
            b = comm.bcast(b, root=0)
            d = comm.bcast(d, root=0)
            diff = comm.bcast(diff, root=0)
            opt_bw = comm.bcast(opt_bw, root=0)
            opt_tau = comm.bcast(opt_tau, root=0)
        else:
            if score_b <= score_d:
                opt_score = score_b
                opt_bw = b
                opt_tau = tau_b
                c = d
                d = b
                b = a + delta * np.abs(c - a)
            else:
                opt_score = score_d
                opt_bw = d
                opt_tau = tau_d
                a = b
                b = d
                d = c - delta * np.abs(c - a) 
            diff = score_b - score_d
            if verbose:
                print('bw: ', opt_bw, ', tau: ', opt_tau, ', score: ', opt_score)
    return opt_bw, opt_tau



import numpy as np
from scipy.spatial.distance import cdist
class GTWRKernel(object):

    def __init__(self, i, coords, t, bw=None, tau=None, fixed=True, 
                 function='triangular', eps=1.0000001):
    
        self.coords = coords
        self.t = t
        self.tau = tau
        if self.tau == 0:
            self.coords_new = coords
        else:
            self.coords_new = np.hstack([coords, (np.sqrt(tau) * t)])
        self.dvec = cdist([self.coords_new[i]], self.coords_new).reshape(-1)
        self.function = function

        if fixed:
            self.bandwidth = float(bw)
        else:
            self.bandwidth = np.partition(
                self.dvec,
                int(bw) - 1)[int(bw) - 1] * eps  #partial sort in O(n) Time
        self.kernel = self._kernel_funcs(self.dvec / self.bandwidth)

        if self.function == "bisquare":  #Truncate for bisquare
            self.kernel[(self.dvec >= self.bandwidth)] = 0

    def _kernel_funcs(self, zs):
        # functions follow Anselin and Rey (2010) table 5.4
        if self.function == 'triangular':
            return 1 - zs
        elif self.function == 'uniform':
            return np.ones(zs.shape) * 0.5
        elif self.function == 'quadratic':
            return (3. / 4) * (1 - zs**2)
        elif self.function == 'quartic':
            return (15. / 16) * (1 - zs**2)**2
        elif self.function == 'gaussian':
            return np.exp(-0.5 * (zs)**2)
        elif self.function == 'bisquare':
            return (1 - (zs)**2)**2
        elif self.function == 'exponential':
            return np.exp(-zs)
        else:
            print('Unsupported kernel function', self.function)
            
            
class GWRKernel(object):

    def __init__(self, i, coords, bw=None, fixed=True, 
                 function='triangular', eps=1.0000001):
    
        self.coords = coords
        self.dvec = cdist([self.coords[i]], self.coords).reshape(-1)
        self.function = function

        if fixed:
            self.bandwidth = float(bw)
        else:
            self.bandwidth = np.partition(
                self.dvec,
                int(bw) - 1)[int(bw) - 1] * eps  #partial sort in O(n) Time
        
        self.kernel = self._kernel_funcs(self.dvec / self.bandwidth)

        if self.function == "bisquare":  #Truncate for bisquare
            self.kernel[(self.dvec >= self.bandwidth)] = 0

    def _kernel_funcs(self, zs):
        # functions follow Anselin and Rey (2010) table 5.4
        if self.function == 'triangular':
            return 1 - zs
        elif self.function == 'uniform':
            return np.ones(zs.shape) * 0.5
        elif self.function == 'quadratic':
            return (3. / 4) * (1 - zs**2)
        elif self.function == 'quartic':
            return (15. / 16) * (1 - zs**2)**2
        elif self.function == 'gaussian':
            return np.exp(-0.5 * (zs)**2)
        elif self.function == 'bisquare':
            return (1 - (zs)**2)**2
        elif self.function == 'exponential':
            return np.exp(-zs)
        else:
            print('Unsupported kernel function', self.function)            
            
            

         
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            


import setuptools

setuptools.setup(
    name="gtwr",
    version="1.0.0",
    long_description = 'To fit geographically and temporally weighted regression model',
    author="Kun Sun",
    author_email="849024477@qq.com",
    packages=['gtwr']
    )
          
        
            



import numpy as np

def get_AICc(gtwr, mpi = False):
    """
    Get AICc value
    
    Gaussian: p61, (2.33), Fotheringham, Brunsdon and Charlton (2002)
    
    GWGLM: AICc=AIC+2k(k+1)/(n-k-1), Nakaya et al. (2005): p2704, (36)

    """
    if mpi:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0:
            n = gtwr.n
            k = gtwr.tr_S
            
            aicc = get_AIC(gtwr) + 2.0 * k * (k + 1.0) / (n - k - 1.0)
            return aicc
    else:
        n = gtwr.n
        k = gtwr.tr_S
        
        aicc = get_AIC(gtwr) + 2.0 * k * (k + 1.0) / (n - k - 1.0)
        return aicc


def get_AIC(gtwr, mpi = False):
    """
    Get AIC calue

    Gaussian: p96, (4.22), Fotheringham, Brunsdon and Charlton (2002)

    GWGLM:  AIC(G)=D(G) + 2K(G), where D and K denote the deviance and the effective
    number of parameters in the model with bandwidth G, respectively.
    
    """
    if mpi:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0:
            k = gtwr.tr_S
            aic = -2.0 * gtwr.llf + 2.0 * (k + 1)
            
            return aic
    else:
            
        k = gtwr.tr_S
    
        aic = -2.0 * gtwr.llf + 2.0 * (k + 1)
        
        return aic


def get_BIC(gtwr, mpi = False):
    """
    Get BIC value

    Gaussian: p61 (2.34), Fotheringham, Brunsdon and Charlton (2002)
    BIC = -2log(L)+klog(n)

    GWGLM: BIC = dev + tr_S * log(n)

    """
    if mpi:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0:
        
            n = gtwr.n  # (scalar) number of observations
            k = gtwr.tr_S
            
            bic = -2.0 * gtwr.llf + (k + 1) * np.log(n)
            return bic
    else:
        n = gtwr.n  # (scalar) number of observations
        k = gtwr.tr_S
        
        bic = -2.0 * gtwr.llf + (k + 1) * np.log(n)
        return bic


def get_CV(gtwr, mpi = False):
    """
    Get CV value

    Gaussian only

    Methods: p60, (2.31) or p212 (9.4)
    Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.
    Modification: sum of residual squared is divided by n according to GWR4 results

    """
    if mpi:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0: 
            cv = gtwr.aa / gtwr.n
            return cv
    else:
        cv = gtwr.aa / gtwr.n
        return cv


def corr(cov, mpi = False):
    if mpi:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        if rank == 0: 
            invsd = np.diag(1 / np.sqrt(np.diag(cov)))
            cors = np.dot(np.dot(invsd, cov), invsd)
            return cors
    else:
        invsd = np.diag(1 / np.sqrt(np.diag(cov)))
        cors = np.dot(np.dot(invsd, cov), invsd)
        return cors